RGM = {Build = {}, Data = {}, Extra = {}, Ignore = {}, Sessions = {}, Share = {}}

function RGM.Build.Engine()
	local Gsub, Lower, Match, String, Upper = string.gsub, string.lower, string.match, string.format, string.upper
	local Register, Inspect, Number = Command.Event.Attach, Inspect, tonumber
	local Insert, Remove = table.insert, table.remove
	local User = Inspect.Unit.Lookup("player")
	local Next, Pairs = next, pairs
	local T, F = true, false
	local E, O = RGM

	function E.Ability(Input, Type, Unit)
		local Input, Pet = Input, (Type == "Given" and Input.caster) and (Input.caster == Unit.Pet or E.Units({[Input.caster] = "Pet"})) or (Type == "Taken" and Input.target) and (Input.target == Unit.Pet or E.Units({[Input.target] = "Pet"}))
		local Info, Name = {Absorbed = {}, Avoided = {}, Blocked = {}, Damage = {}, Heal = {}, Intercepted = {}, Missed = {}, Overheal = {}, Overkill = {}}, Input.abilityName
		local ID = String("%0s (%0s)", Input.abilityName, (Input.converted and "Absorb" or (Input.heal or Input.overheal) and "Heal" or "Damage"))
		local Ability = Unit.Abilities[ID]
		
		if Pet then
			ID = String("%0s (%0s)", ID, "Pet")
			Ability = Unit.Abilities[ID]
		end
		
		if not Ability then
			if not Input.converted then
				if not Input.heal and not Input.overheal then
					Name = String("%0s (%0s)", Name, (Input.type or "Physical"))
				else
					Name = String("%0s (Heal)", Name)
				end
			else
				Name = String("%0s (Absorb)", Name)
			end
			
			if Pet then
				Name = String("%0s (Pet)", Name)
			end
			
			Unit.Abilities[ID] = {[Type] = Info, Tally = {}, Tooltip = Name, Update = Input.damageAvoided}
			Ability = Unit.Abilities[ID]
		end
		
		if Ability then
			Ability.Tally[Type] = (Ability.Tally[Type] or 0) + 1
		end
		
		if Ability.Update and (Input.damage or Input.damageAbsorbed) then
			Ability.Tooltip, Ability.Update = String("%0s (%0s)%0s", Name, (Input.type or "Physical"), (Pet and " (Pet)" or ""))
		end
		
		if Ability and not Ability[Type] then
			Ability[Type] = Info
		end
		
		return Ability[Type]
	end

	function E.Absorb(Debug, Input)
		local Input = E.Private(Input) if not Input then return end
		
		local Ran = O
		
		for k,v in Pairs(Input) do
			local Unit = (E.Extra[k] or {Absorb = {}})
			
			for i = #Unit.Absorb, 1, -1 do
				local Absorb = Unit.Absorb[i]
				
				if v and v > Unit.Value and not Absorb.Value and not Ran then
					Absorb.Value, Ran = (v - Unit.Value), 0
				end
			end
			
			Unit.Value, Ran = (v or 0), O
		end
	end

	function E.Add(Debug, Unit, Input)
		local Input = E.Private(Input) if not Input or not Unit then return end
		
		local Detail, Lookup = Inspect.Buff.Detail, Inspect.Unit.Lookup(Unit)
		local ID, Ran = E.Extra[Unit], O
		
		if not E.Extra[Unit] then
			local Blocked = Lookup and (Match(Lookup, "focus") or Match(Lookup, "mouseover") or Match(Lookup, "target"))
			
			if not Blocked and Lookup then
				if Match(Lookup, "group") or Match(Lookup, "player") then
					E.Extra[Unit] = {Absorb = {}, Value = 0}
					ID = E.Extra[Unit]
				end
			end
		end
		
		if not E.Extra[Unit] then return end
		
		for k,v in Pairs(Input) do
			if E.Ignore[v] then break end
			
			local Begin, Duration, Find = O, O, O
			local Info = Detail(Unit, k)
			
			if Info then
				for k,v in Pairs(E.Filter.Convert) do
					Find = (Match(Info.description or "", v) or Match(Info.description or "", Gsub(v, "%u", Lower)) or Find)
				end
				
				Begin, Duration = (Info.begin or 0), (Info.duration or 0 - 0.04)
			end
			
			if Find and Duration <= 80 then
				Ran = (Insert(ID.Absorb, {Caster = Info.caster, End = Begin + Duration, Buff = k, Name = Info.name}) or 0)
			end
			
			if Info and not Ran then
				E.Ignore[v] = 0
			end
			
			Ran = O
		end
	end

	function E.Blocked(Debug, Input)
		local Input = E.Private(Input) if not Input then return end
		
		for k,v in Pairs(Input) do
			if E.Data[k] and E.Instance then
				E.Data[k].Blocked = v
			end
		end
	end

	function E.Clean(Input)
		if not Input then return end
		
		local Info = {}
		
		if Info then
			E.Data, E.Ignore, E.Share = {}, {}, {}
			E.Active, E.Instance = O, O
		end
		
		if Input.Log and not Next(Input.Log) then
			Input.Log = O
		end
		
		for k,v in Pairs(Input) do
			local Abilities = (v.Abilities or Info)
			
			for k,v in Pairs(Abilities) do
				local Ability = (v or Abilities[k])
				
				for k,v in Pairs(Ability) do
					local Params = (k == "Given" or k == "Taken")
					local Data = (v or Ability[k])
					
					if Params then
						for k,v in Pairs(Data) do
							if Data[k] and not Next(v) then
								Data[k] = O
							end
						end
						
						if not Next(Data) then
							Ability[k] = O
						end
					end
					
					if k and k == "Update" then
						Ability[k] = O
					end
				end
				
				if not v.Given and not v.Taken then
					Abilities[k] = O
				end
			end
			
			if v.Log then
				if not v.Name or not v.Abilities then
					Input[k] = O
				end
				
				if not Next(v.Log) then
					v.Log = O
				end
			end
			
			v.Blocked = O
			v.Combat = O
		end
		
		return Input
	end

	function E.Clear(Debug, Input)
		local Input = E.Private(Input) if not Input then return end
		
		local Data = {}
		
		for k,v in Pairs(Input) do
			if E.Data[k] and E.Data[k].Combat and E.Data[k].Role and not v then
				Data[k] = F
			end
			
			if E.Extra[k] and not v then
				E.Extra[k] = O
			end
		end
		
		E.Combat(O, Data)
	end

	function E.Combat(Debug, Input)
		local Input = E.Private(Input) if not Input then return end
		
		local Units = Inspect.Unit.List()
		local Combat = O
		
		if not E.Active and Next(E.Data) then
			E.Data = {}
		end
		
		for k,v in Pairs(Input) do
			if not Next(E.Data) and not E.Active and Units and v then
				E.Group(Units)
			end
			
			if E.Data[k] and not E.Data[k].Combat and v then
				E.Data[k].Combat = 0
			end
			
			if E.Data[k] and E.Data[k].Combat and not v then
				E.Data[k].Combat = O
			end
		end
		
		for k,v in Pairs(E.Data) do
			local Player = (k == User or E.Instance and v.Role)
			
			if Input[v.Pet] and not v.Combat then
				v.Combat = 0
			end
			
			if not Units[k] and v.Combat then
				v.Combat = O
			end
			
			if Player and v.Combat then
				Combat = 0
			end
		end
		
		if Combat and not E.Active then
			Command.Message.Broadcast("yell", O, "RGM", String("#<%0s>", User))
		end
		
		if not Combat and E.Active then
			E.Sessions[#E.Sessions + 1] = E.Clean(E.Data)
			E.Interface[1].Session(E.Interface[2] or 0)
		end
		
		E.Active = Combat
	end

	function E.Command(Debug, Input)
		if not Input or not Match(Input, "%a+") then return end
		
		local Global, One, Two = RGM_Global, E.Interface[1], E.Interface[2]
		
		if Match(Input, "Density") or Match(Input, "density") then
			local Density = Number(Match(Input, "%d.%d") or Match(Input, "%d"))
			
			if Density and (Density > 0 and Density <= 1) then
				Global[1].Density = Density
				Global[2].Density = Density
			end
			
			local Run = (One and One.Density() or Two and Two.Density())
		end
		
		if Input == "Toggle" or Input == "toggle" then
			Global[1].Toggle = not Global[1].Toggle
			Global[2].Toggle = not Global[2].Toggle
			
			if One and not Global[1].Toggle then
				One.Main:SetVisible(Global[1].Show)
			end
			
			if Two and not Global[2].Toggle then
				Two.Main:SetVisible(Global[2].Show)
			end
			
			if One and Global[1].Toggle then
				One.Main:SetVisible(F)
			end
			
			if Two and Global[2].Toggle then
				Two.Main:SetVisible(F)
			end
		end
		
		if Input == "Reset" or Input == "reset" then
			local Run = (One and One.Reset() or Two and Two.Reset())
		end
		
		if Input == "Over" or Input == "over" then
			Global[1].Over = not Global[1].Over
			Global[2].Over = not Global[2].Over
		end
	end

	function E.Converter(Input)
		local Input, Unit = Input, (E.Extra[Input.target] or {Absorb = {}})
		local Integer, Time = math.floor, Inspect.Time.Frame()
		local Heal = (Input.damageAbsorbed or 0)
		local Params = {converted = 0}
		local Ran = O
		
		for k,v in Pairs(Unit.Absorb) do
			if Integer(Time) >= Integer(v.End + 0.04) then
				Remove(Unit.Absorb, k)
			end
		end
		
		for i = #Unit.Absorb, 1, -1 do
			local v = (not Ran and i and Unit.Absorb[i])
			local Caster = (v and E.Data[v.Caster])
			
			if not Caster and v then
				Caster = E.Units({[v.Caster] = "Return"})
			end
			
			if Caster then
				Params.targetName = Input.targetName
				Params.casterName = Caster.Name
				Params.target = Input.target
				Params.abilityName = v.Name
				Params.caster = v.Caster
				
				if not Params.heal and v.Value and v.Value > 0 and v.Value >= Heal then
					v.Value = v.Value - Heal
					Params.heal = Heal
					Ran = Params.heal
				end
				
				if not Params.heal and v.Value and v.Value > 0 and v.Value < Heal then
					Ran = not Remove(Unit.Absorb, i)
					Params.heal = v.Value
					Heal = Heal - v.Value
				end
				
				if v and Params.heal then
					E.Gather(Params)
				end
			end
			
			Params = {converted = 0}
		end
	end

	function E.Count(Data, Event, Input, Name, Type, Filter)
		local Input, Convert = Input, (Input.damageAbsorbed or Input.damageIntercepted or Input.damageBlocked)
		local Filter, Name = (E.Filter[Type] and E.Filter[Type][Filter]), Name or "Unknown"
		
		if not Filter or Filter[Input.abilityName] then return end
		
		if Type == "Taken" and Convert then
			Input.damage = Input.taken
		end
		
		if Event == "damageAvoided" then
			Input.damageAvoided = O
		end
		
		if Name and not Data.Units then
			Data.Units = {}
		end
		
		if not Input.crit then
			Data["Normal Count"] = (Data["Normal Count"] or 0) + 1
			
			if Input[Event] then
				if (not Data["Normal Max"] or Data["Normal Max"] < Input[Event]) then
					Data["Normal Max"] = Input[Event]
				end
				
				if (not Data["Normal Min"] or Data["Normal Min"] > Input[Event]) then
					Data["Normal Min"] = Input[Event]
				end
				
				Data["Normal Total"] = (Data["Normal Total"] or 0) + Input[Event]
			end
		end
		
		if Input.crit then
			Data["Critical Count"] = (Data["Critical Count"] or 0) + 1
			
			if Input[Event] then
				if (not Data["Critical Max"] or Data["Critical Max"] < Input[Event]) then
					Data["Critical Max"] = Input[Event]
				end
				
				if (not Data["Critical Min"] or Data["Critical Min"] > Input[Event]) then
					Data["Critical Min"] = Input[Event]
				end
				
				Data["Critical Total"] = (Data["Critical Total"] or 0) + Input[Event]
			end
		end
		
		if Name then
			Data.Units[Name] = (Data.Units[Name] or 0) + (Input[Event] or 1)
		end
	end

	function E.Damage(Debug, Input)
		E.Processor(O, Input)
	end

	function E.Death(Debug, Input)
		local Input = E.Private(Input) if not Input then return end
		
		local One, Two = E.Interface[1], E.Interface[2]
		local Unit = E.Data[Input.target]
		
		if Unit and Unit.Role and One and not One.Temp.Index and One.Log:GetVisible() then
			One = One.Log.Update(E.Data.Log)
		end
		
		if Unit and Unit.Role and Two and not Two.Temp.Index and Two.Log:GetVisible() then
			Two = Two.Log.Update(E.Data.Log)
		end
	end

	function E.Dodge(Debug, Input)
		E.Processor(0, Input)
	end

	function E.Enter(Input)
		E.Combat(Input, {[User] = T})
	end

	function E.Exit(Input)
		E.Combat(Input, {[User] = F})
	end

	function E.Gather(Input, Sync)
		local Caster, Target = {Abilities = {}}, {Abilities = {}}
		local Input, Given, Taken = Input, 0, 0
		
		if Caster then
			Caster = (E.Search(Input.casterName, Input.caster) or Given and Caster)
		end
		
		if Target then
			Target = (E.Search(Input.targetName, Input.target) or Taken and Target)
		end
		
		if not Input.converted and not Sync then
			if Caster.Blocked and E.Sync(Caster.Blocked, Input, Input.caster) then
				return
			end
			
			if Target.Blocked and E.Sync(Target.Blocked, Input, Input.target) then
				return
			end
		end
		
		if Caster.Abilities then
			Input.casterName = (Caster.Name or Input.casterName)
			Given = E.Ability(Input, "Given", Caster)
		end
		
		if Target.Abilities then
			Input.targetName = (Target.Name or Input.targetName)
			Taken = E.Ability(Input, "Taken", Target)
		end
		
		if Input.damage then
			E.Count(Given.Damage, "damage", Input, Target.Name, "Given", "Damage")
			E.Count(Taken.Damage, "damage", Input, Caster.Name, "Taken", "Damage")
		end
		
		if Input.damageAbsorbed then
			E.Count(Given.Absorbed, "damageAbsorbed", Input, Target.Name, "Given", "Absorb")
			E.Count(Taken.Absorbed, "damageAbsorbed", Input, Caster.Name, "Taken", "Absorb")
		end
		
		if Input.damageAvoided then
			E.Count(Given.Missed, "damageAvoided", Input, Target.Name, "Given", "Miss")
			E.Count(Taken.Avoided, "damageAvoided", Input, Caster.Name, "Taken", "Avoid")
		end
		
		if Input.damageBlocked then
			E.Count(Given.Blocked, "damageBlocked", Input, Target.Name, "Given", "Block")
			E.Count(Taken.Blocked, "damageBlocked", Input, Caster.Name, "Taken", "Block")
		end
		
		if Input.damageIntercepted then
			E.Count(Given.Intercepted, "damageIntercepted", Input, Target.Name, "Given", "Intercept")
			E.Count(Taken.Intercepted, "damageIntercepted", Input, Caster.Name, "Taken", "Intercept")
		end
		
		if Input.heal then
			E.Count(Given.Heal, "heal", Input, Target.Name, "Given", "Heal")
			E.Count(Taken.Heal, "heal", Input, Caster.Name, "Taken", "Heal")
		end
		
		if Input.overheal then
			E.Count(Given.Overheal, "overheal", Input, Target.Name, "Given", "Overheal")
			E.Count(Taken.Overheal, "overheal", Input, Caster.Name, "Taken", "Overheal")
		end
		
		if Input.overkill then
			E.Count(Given.Overkill, "overkill", Input, Target.Name, "Given", "Overkill")
			E.Count(Taken.Overkill, "overkill", Input, Caster.Name, "Taken", "Overkill")
		end
		
		if Target.Role then
			if Input.damage or Input.damageAbsorbed or Input.heal then
				E.Log(Input, Target)
			end
			
			if Input.damageAbsorbed then
				E.Converter(Input)
			end
		end
	end

	function E.Group(Input)
		local Unit, Zone = Inspect.Unit.Detail, Inspect.Zone.Detail
		local Pet = Inspect.Unit.Lookup(User .. ".pet")
		local Player = Unit(User)
		local Instance = O
		
		if Player then
			local Zone = (Player.zone and Zone(Player.zone) or O)
			
			if Zone and Zone.type == "instance" then
				Instance = 0
			end
		end
		
		for k,v in Pairs(Input) do
			local Inspect = (not Instance and Unit(k))
			
			if Inspect and (Inspect.tier == "raid" or Inspect.level == "??") then
				Instance = not Match(Inspect.name, "Dummy")
			end
		end
		
		if not Instance and (Input[Pet] or Input[User]) then
			Input = {[User] = 0}
		end
		
		if Instance or Input[Pet] or Input[User] then
			E.Units(Input)
		end
		
		E.Instance = Instance
	end

	function E.Heal(Debug, Input)
		E.Processor(O, Input)
	end

	function E.Log(Input, Unit)
		local Input, Convert = Input, (Input.damageAbsorbed or Input.damageIntercepted or Input.damageBlocked)
		local Session = (E.Data.Time and E.Data or Input)
		local One, Two = E.Interface[1], E.Interface[2]
		local Damage, Heal, Time = 0, 0, 0
		
		if Convert and Input then
			Input.damage = Input.taken
		end
		
		if Unit then
			Damage = String("%0s's %0s Hits %0s For %0s %0s Damage.", Input.casterName, Input.abilityName .. (Input.crit and " Critically" or ""), Input.targetName, (Input.damage or 0), (Input.type or "Physical"))
			Heal = String("%0s's %0s Heals %0s For %0s.", Input.casterName, Input.abilityName .. (Input.crit and " Critically" or ""), Input.targetName, (Input.heal or 0))
		end
		
		if Input then
			Time = (Time and Session.Time and One.Duration(Session) or "00:00")
			Session.Log, Unit.Log = (Session.Log or {}), (Unit.Log or {})
		end
		
		if Input.damage or Input.damageAbsorbed then
			local Intercepted = (Input.damageIntercepted and String(" (%0s Intercepted)", Input.damageIntercepted) or "")
			local Absorbed = (Input.damageAbsorbed and String(" (%0s Absorbed)", Input.damageAbsorbed) or "")
			local Blocked = (Input.damageBlocked and String(" (%0s Blocked)", Input.damageBlocked) or "")
			local Overkill = (Input.overkill and String(" (%0s Overkill)", Input.overkill) or "")
			
			Damage = String("[%0s] %0s%0s%0s%0s%0s", Time, Damage, Absorbed, Blocked, Intercepted, Overkill)
			
			Insert(Unit.Log, String("<font color=%0s#D20000%0s> [Damage] %0s </font>", '"', '"', Damage))
		end
		
		if Input.heal then
			Heal = String("[%0s] %0s%0s", Time, Heal, (Input.overheal and String(" (%0s Overheal)", Input.overheal) or ""))
			
			Insert(Unit.Log, String("<font color=%0s#04B404%0s> [Heal] %0s </font>", '"', '"', Heal))
		end
		
		if Input.damage and Input.overkill then
			Insert(Session.Log, String("<font color=%0s#F28000%0s> %0s </font>", '"', '"', Damage))
			
			if Input.target == User then
				print(Damage)
			end
		end
		
		if One and not One.Temp.Index and One.Temp.Log and One.Temp.Log[2] == Unit.Name then
			One.Log.Update(Unit.Log, Unit.Name)
		end
		
		if Two and not Two.Temp.Index and Two.Temp.Log and Two.Temp.Log[2] == Unit.Name then
			Two.Log.Update(Unit.Log, Unit.Name)
		end
		
		if Unit and #Unit.Log >= 28 then
			Remove(Unit.Log, 1)
		end
	end

	function E.Miss(Debug, Input)
		E.Processor(0, Input)
	end

	function E.Part(Debug, Input)
		local Input = E.Private(Input) if not Input then return end
		
		local Data = {}
		
		for k,v in Pairs(Input) do
			if E.Data[k] and E.Data[k].Combat and E.Data[k].Role then
				Data[k] = F
			end
		end
		
		E.Combat(O, Data)
	end

	function E.Private(Input)
		if not Input or not Next(Input) then return end
		
		local Output = {}
		
		for k,v in Pairs(Input) do
			Output[k] = v
		end
		
		return Output
	end

	function E.Processor(Avoid, Input)
		local Input = E.Private(Input) if not Input then return end
		
		local Convert = (Input.damageAbsorbed or Input.damageIntercepted or Input.damageBlocked)
		local Event = (Avoid or Input.damage or Input.damageAbsorbed)
		local Self = (Input.caster == Input.target)
		local Damage = Input.damage
		
		if Convert and Self or Event and Self then
			return
		end
		
		if E.Active then
			Input.damageIntercepted = (Input.damageIntercepted and Input.damageIntercepted < 0 and -Input.damageIntercepted or Input.damageIntercepted)
			Input.ability, Input.abilityNew, Input.damageAvoided, Input.type = O, O, Avoid, (Input.type and Gsub(Input.type, "^%l", Upper))
			Input.damageAbsorbed = (Input.damageAbsorbed and Input.damageAbsorbed < 0 and -Input.damageAbsorbed or Input.damageAbsorbed)
		else
			return
		end
		
		if Convert then
			Input.damage = (Input.damage or 0) + (Input.damageAbsorbed or 0) + (Input.damageIntercepted or 0) + (Input.damageBlocked or 0)
			Input.taken = Damage
		end
		
		if not Input.abilityName or Match(Input.abilityName, "attack") or not Match(Input.abilityName, "%a+") then
			Input.abilityName = "Auto Attack"
		end
		
		if not Input.casterName or not Match(Input.casterName, "%a+") then
			Input.casterName = "Unknown"
		end
		
		if not Input.targetName or not Match(Input.targetName, "%a+") then
			Input.targetName = "Unknown"
		end
		
		if Input.damageAbsorbed and not Input.damage then
			Input.damage = 0
		end
		
		E.Gather(Input)
	end

	function E.Receive(Debug, From, Type, Channel, Addon, Data)
		if not Addon or not Data or not (Addon == "RGM") then return end
		
		local Data, Events = Data, {damage = 0, damageAbsorbed = 0, damageAvoided = 0, damageBlocked = 0, damageIntercepted = 0, heal = 0, overheal = 0, overkill = 0, taken = 0}
		local Unit = Match(Data, "#<(.-)>")
		local Info = {}
		
		if Data and Info then
			Info.abilityName = Match(Data, "abilityName<(.-)>")
			Info.casterName = Match(Data, "casterName<(.-)>")
			Info.targetName = Match(Data, "targetName<(.-)>")
			Info.caster = Match(Data, "caster<(.-)>")
			Info.target = Match(Data, "target<(.-)>")
			Info.type = Match(Data, "type<(.-)>")
			Info.crit = Match(Data, "crit<")
		end
		
		for k,v in Pairs(Events) do
			Info[k] = Number(Match(Data, String("%0s<(.-)>", k)))
		end
		
		if Unit and not E.Share[Unit] then
			E.Share[Unit] = 0
		end
		
		if Next(Info) and E.Active then
			E.Gather(Info, 0)
		end
	end

	function E.Remove(Debug, Unit, Input)
		local Input = E.Private(Input) if not Input or not Unit then return end
		
		local Integer, Time = math.floor, Inspect.Time.Frame()
		local Data = E.Extra[Unit]
		
		if Data and Input then
			for k,v in Pairs(Data.Absorb) do
				if Integer(Time) >= Integer(v.End) then
					Remove(Data.Absorb, k)
				end
			end
			
			Data = O
		end
	end

	function E.Resist(Debug, Input)
		E.Processor(0, Input)
	end

	function E.Search(Name, Unit)
		if Unit and E.Data[Unit] then return E.Data[Unit] end
		
		local Data = E.Data
		
		for k,v in Pairs(Data) do
			if Name and v.Name and v.Name == Name and not v.Player then
				return Data[k]
			end
			
			if Unit and v.Pet and v.Pet == Unit then
				return Data[k]
			end
		end
		
		return (Unit and E.Units({[Unit] = "Return"}) or O)
	end

	function E.Send(Input, Unit, Name)
		if not Name or Unit == User then return end
		
		local Send = Command.Message.Send
		local Input, Data = Input, ""
		
		for k,v in Pairs(Input) do
			Data = String("%0s%0s<%0s>", Data, k, v)
		end
		
		Send(Name, "RGM", Data, tostring)
	end

	function E.Sync(Blocked, Input, Unit)
		if not E.Instance or not Unit then return end
		
		local Input, Event = Input, (Input.damage or Input.damageAbsorbed or Input.damageAvoided)
		local Owner = Inspect.Unit.Lookup(Unit .. ".owner")
		
		for k,v in Pairs(E.Data) do
			if E.Share[k] and Event and Input.target == Unit and Unit == User and v.Blocked then
				E.Send(Input, k, v.Name)
			end
			
			if E.Share[k] and Input.caster == Unit and Unit == User and v.Blocked then
				E.Send(Input, k, v.Name)
			end
		end
		
		if Blocked and (E.Share[Unit] or E.Share[Owner]) then
			return Unit
		end
	end

	function E.Units(Input)
		local Detail, Lookup = Inspect.Unit.Detail, Inspect.Unit.Lookup
		local Owner, Unit = O, O
		
		for k,v in Pairs(Input) do
			if not Match(k, "%a%w+") then break end
			
			local i = (Lookup(k .. ".owner") or k)
			
			if Detail and not E.Data[i] then
				Unit = Detail(i)
			end
			
			if Unit then
				Unit.name = (Match(Unit.name or "", "%a+") and Unit.name)
			end
			
			if Unit and (Unit.tier == "raid" or Unit.level == "??") then
				E.Instance = not Match(Unit.name or "", "Dummy")
			end
			
			if Unit then
				local Info = (Unit.alliance and Unit.nameSecondary)
				
				if Info then
					local Name = Unit.nameSecondary
					
					for k,v in Pairs(E.Data) do
						if v.Name and v.Player and Match(Name, v.Name) then
							Owner, v.Pet = k, Unit.id
						end
					end
					
					if not Owner and not (v == "Pet") then
						return
					end
				end
				
				if Owner and E.Data[Owner] then
					return E.Data[Owner]
				end
			end
			
			if not Owner and k and v == "Pet" then
				return Lookup(k .. ".owner")
			end
			
			if not E.Data[i] and Unit then
				E.Data[i] = {Abilities = {}, Log = {}}
				
				E.Data[i].Calling = Unit.calling
				E.Data[i].Combat = Unit.combat
				E.Data[i].Level = Unit.level
				E.Data[i].Name = Unit.name
				E.Data[i].Player = Unit.player
				E.Data[i].Relation = Unit.relation
				
				E.Data[i].Pet = Lookup(i .. ".pet")
				E.Data[i].Blocked = Unit.blocked
				E.Data[i].Role = Unit.role
			end
			
			if v and v == "Return" then
				return E.Data[i]
			end
			
			Owner = O
			Unit = O
		end
	end

	function E.Update(Input)
		if not E.Active then return end
		
		local Time = Inspect.Time.Real()
		
		if E.Data.Time then
			E.Data.Time.Duration = Time - E.Data.Time.Start
		else
			E.Data.Time = {Duration = 0, Start = Time}
		end
		
		E.Interface[1].Current(E.Interface[2])
	end

	if Next(E) then
		Register(Command.Slash.Register("RGM"), E.Command, "Slash")
		
		Register(Event.Buff.Add, E.Add, "Add Buff")
		Register(Event.Buff.Remove, E.Remove, "Remove Buff")
		Register(Event.Combat.Damage, E.Damage, "Damage")
		Register(Event.Combat.Death, E.Death, "Deaths")
		Register(Event.Combat.Dodge, E.Dodge, "Dodge")
		Register(Event.Combat.Heal, E.Heal, "Heal")
		Register(Event.Combat.Miss, E.Miss, "Miss")
		Register(Event.Combat.Resist, E.Resist, "Resist")
		Register(Event.Message.Receive, E.Receive, "Sync")
		Register(Event.System.Secure.Enter, E.Enter, "Enter")
		Register(Event.System.Secure.Leave, E.Exit, "Exit")
		Register(Event.System.Update.Begin, E.Update, "Update")
		Register(Event.Unit.Availability.Partial, E.Part, "Part")
		Register(Event.Unit.Detail.Absorb, E.Absorb, "Absorb")
		Register(Event.Unit.Detail.Blocked, E.Blocked, "Blocked")
		Register(Event.Unit.Detail.Combat, E.Combat, "Combat")
		Register(Event.Unit.Remove, E.Clear, "Remove Unit")
	end

	RGM.User = User
end